#ifndef PINBALL_ENTITIES_H
#define PINBALL_ENTITIES_H

#include "stdint-gcc.h"
#include "utilities.h"
#include "vector2D.h"
#include <num/num.h>

#include <vector>

class Ball {
public:
  Ball(libnum::num32 radius, libnum::num32 mass, Vector2D pos, Vector2D vel,
       libnum::num32 restitution) {
    this->radius = radius;
    this->mass = mass;
    this->pos = pos.Clone();
    this->vel = vel.Clone();
    this->restitution = restitution;
  }

  ~Ball() {}

  void Simulate(libnum::num32 dt, Vector2D gravity) {
    this->vel.Add(gravity, dt);
    this->pos.Add(this->vel, dt);
  }

  libnum::num32 radius, mass, restitution;
  Vector2D pos, vel;
};


class Obstacle {
public:
  Obstacle(libnum::num32 radius, Vector2D pos, libnum::num32 pushVel) {
    this->radius = radius;
    this->pos = pos.Clone();
    this->pushVel = pushVel;
  }

  Obstacle() {}

  libnum::num32 radius, pushVel;
  Vector2D pos;
};


class Flipper {
public:
  Flipper(libnum::num32 radius, Vector2D pos, libnum::num32 length,
          libnum::num32 restAngle, libnum::num32 maxRotation,
          libnum::num32 angularVelocity, libnum::num32 restitution) {
    this->radius = radius;
    this->pos = pos.Clone();
    this->length = length;
    this->restAngle = restAngle;
    this->maxRotation = ABS(maxRotation);
    this->sign = SIGN(maxRotation);
    this->angularVelocity = angularVelocity;
    this->rotation = libnum::num32(0);
    this->currentAngularVelocity = libnum::num32(0);
    this->touchIdentifier = libnum::num32(-1);
  }

  void Simulate(libnum::num32 dt) {
    libnum::num32 prevRotation = this->rotation;
    bool pressed = this->touchIdentifier >= 0;
    if (pressed)
      this->rotation = MIN(this->rotation + dt * this->angularVelocity, this->maxRotation);
    else
      this->rotation = MAX(this->rotation - dt * this->angularVelocity, 0.0);
    this->currentAngularVelocity = this->sign * (this->rotation - prevRotation) / dt;
  }

  bool Select(Vector2D pos)
  {
    Vector2D D;;
    D.SubtractVectors(this->pos, pos);
    bool test = (D.Length() < this->length);
    return test;
  }

  Vector2D getTip()
  {
    libnum::num32 angle = this->restAngle + this->sign * this->rotation;
    Vector2D D( COS(angle) ,  SIN(angle) );

    Vector2D tip = this->pos.Clone();
    tip.Add(D, this->length);
    return tip;
  }

  libnum::num32 radius;
  Vector2D pos;
  libnum::num32 length;
  libnum::num32 restAngle;
  libnum::num32 maxRotation;
  libnum::num32 sign;
  libnum::num32 angularVelocity;
  libnum::num32 rotation;
  libnum::num32 currentAngularVelocity;
  libnum::num32 touchIdentifier;
};


struct Scene
{
    Vector2D gravity;
    libnum::num32 dt;
    uint32_t score;
    bool pause;
    std::vector<Vector2D> borders;
    std::vector<Ball> balls;
    std::vector<Obstacle> obstacles;
    std::vector<Flipper> flippers;
};


#endif